from DTUtils import Logger
from datetime import datetime, timedelta
import requests
import socket
import sys
from requests.auth import HTTPDigestAuth
from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import ConfigException

from collections import defaultdict

import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class mongoDBAtlasRemote(RemoteBasePlugin):


    GROUP_EVENT_TYPES = [ "CREDIT_CARD_ABOUT_TO_EXPIRE", "NO_PRIMARY", "CPS_SNAPSHOT_FALLBACK_FAILED", "CPS_RESTORE_FAILED", 
                         "AWS_ENCRYPTION_KEY_NEEDS_ROTATION", "AZURE_ENCRYPTION_KEY_NEEDS_ROTATION",
                        "GCP_ENCRYPTION_KEY_NEEDS_ROTATION", "HOST_DOWN", "TOO_MANY_ELECTIONS","REPLICATION_OPLOG_WINDOW_RUNNING_OUT",
                        "NDS_X509_USER_AUTHENTICATION_MANAGED_USER_CERTS_EXPIRATION_CHECK", "NDS_X509_USER_AUTHENTICATION_CUSTOMER_CA_EXPIRATION_CHECK", 
                        "CLUSTER_MONGOS_IS_MISSING", "JOINED_GROUP"]

    def initialize(self, **kwargs):
        config = kwargs['config']

        self.json_config = kwargs["json_config"]

        self.logger = Logger(bool(self.config['debug']), __name__)

        # validate configuration items
        self.group_id = config['group_id'].strip()
        if self.group_id == "":
            raise ConfigException("group ID cannot be empty")

        self.api_url = config['api_url'].strip()
        if self.api_url == "" or self.api_url.find("http") == -1:
            raise ConfigException ("api_url invalid or empty")

        self.api_user = config['api_user'].strip()
        if self.api_user == "":
            raise ConfigException ("API user cannot be empty")

        self.api_token = config['api_token'].strip() 
        if self.api_token == "":
            raise ConfigException("API token cannot be empty")

        self.use_proxy = config['use_proxy']
        self.proxies = None

        self.capture_partitions = config['capture_partitions']

        # Prepare proxy config for 'read' function
        if self.use_proxy:
            self.logger.info("Specified proxy will be used")
            self.proxy_conf = config['proxy_conf'].strip()
            if self.proxy_conf == "":
                raise ConfigException ("Proxy address cannot be empty")
            else:
                self.proxies = { 'https' : self.proxy_conf, 'http' : self.proxy_conf}    

        #SSL client certificate (probably self signed)
        verifyCertificate = self.config.get("certificate", False)
        certificatePath = self.config.get("certificate_path","")
        self.verify = ""
        if verifyCertificate is True and certificatePath != "":
            self.verify = certificatePath


        # Retrieve metric keys from plugin.json
        self.args = {}
        self.args['metrics'] = self.initialize_metrics()


        # retrieve the group name, only do this once
        group_info = self.read("/groups/" + str(self.group_id))
        if group_info is not None:
            self.groupName = group_info["name"]
        else:
            raise ConfigException("Cannot find group name")



    def query(self, **kwargs):

        # to avoid timing issues, get the data from 2 minutes ago.
        now = datetime.utcnow()
 
        start_time = now - timedelta(minutes=2, seconds=now.second, microseconds=now.microsecond-1)
        iso_date_start = start_time.isoformat() + "Z"

        end_time = now - timedelta( minutes=1, seconds=now.second, microseconds=now.microsecond-1)
        iso_date_end = end_time.isoformat() + "Z"

        event_info = self.read("/groups/" + str(self.group_id) + "/events?itemsPerPage=500&minDate=" + iso_date_start + "&maxDate=" + iso_date_end )

        eventList = self.processEvent(event_info)

        cluster_info = self.read("/groups/" + str(self.group_id) + "/clusters")
        node_info = self.read("/groups/" + str(self.group_id) + "/processes")

        if cluster_info is not None and node_info is not None:
            self.report_data(self.groupName, cluster_info, node_info, eventList)
        else:
            self.logger.info("no data so skipping")

        
        # check for organisation level events:
        if len(eventList["organisation"]) > 0:
            myGroup = self.topology_builder.create_group("Organization", "Organization")
            myDevice = myGroup.create_device(self.group_id, self.group_id)
            for item in eventList["organisation"]:
                myDevice.report_error_event(item,item)
        


    def report_data(self, groupName, cluster_info, node_info, eventList):

        # iterate through the clusters defined in the group
        for cluster in cluster_info["results"]:

            group_temp = self.topology_builder.create_group(str(cluster["id"]), "Cluster " + str(cluster["name"]))

            self.logger.info("Creating group for cluster" + str(cluster["name"]))

            # add top level cluster properties

            group_temp.report_property("Project Name", str(groupName))
            group_temp.report_property("Cluster Type", str(cluster["clusterType"]))
            if "providerSettings" in cluster:
                group_temp.report_property("Instance Size Name", str(cluster["providerSettings"]["instanceSizeName"]))

            if "replicationFactor" in cluster:
                group_temp.report_property("Replication Factor", str(cluster["replicationFactor"]))

            if "regionName" in cluster:
                group_temp.report_property("Region Name", str(cluster["providerSettings"]["regionName"]))

            self.reportClusterEvents(eventList["clusters"], cluster["name"], group_temp)

            Nodes = self.getNodeNames(cluster)
            # iterates through the hosts in the cluster
            for node_name in Nodes:

                # remove the port
                aliasName = node_name.split(":")[0]

                node_entry = None
                # check that the host is in the hosts list
                for iter in node_info["results"]:
                    if iter["userAlias"] == aliasName:
                        node_entry = iter
                        break
                if node_entry == None:
                    self.logger.info("Cannot find host in process list " + aliasName)
                    continue

                # alias and hostname are not the same so match on alias from the cluster to the process data!
                hostname = iter["hostname"]
                self.logger.info("Creating device " + hostname)
                device_temp = group_temp.create_device(hostname , hostname)
                try:
                    node_ip = socket.gethostbyname(hostname)
                    #self.logger.info("IP:" + node_ip + "|" + hostname)
                except Exception as e:
                    self.logger.info("Cannot resolve hostname for " + hostname)
                    node_ip = ""

                device_temp.add_endpoint(ip=node_ip, port=node_entry["port"], dnsNames=[hostname])

                device_temp.report_property("Version", str(node_entry["version"]))
                device_temp.report_property("Type", str(node_entry["typeName"]))
                if "replicaSetname" in node_entry:
                    device_temp.report_property("ReplicaSetName", str(node_entry["replicaSetName"]))


                #get the metrics for the host
                process_metrics = self.read("/groups/" + str(self.group_id) + "/processes/" + str(node_name) + "/measurements?granularity=PT2M&period=PT4M&itemsPerPage=500")

                if process_metrics is not None and "measurements" in process_metrics:
                    for metric in process_metrics["measurements"]:
                        if str(metric["name"]) in self.args['metrics'].keys():
                            parsed_metrics = self.parse_metrics(metric["dataPoints"])
                            if parsed_metrics is not None:
                                source = self.args['metrics'][str(metric['name'])]
                                # source tag used to dynamically call a transformation function if necessary
                                if source != "":
                                    try:
                                        myFunction = getattr(self,source)
                                        parsed_metrics = myFunction(parsed_metrics)

                                    except Exception as e:
                                        self.logger.info("cannot process source:" + str(e))
                               

                                #self.logger.debug("Writing metric " + str(metric["name"]) + ":" + str(parsed_metrics))
                                device_temp.absolute(key=str(metric["name"]), value=str(parsed_metrics))
                            else:
                               self.logger.debug("MISSING: " + str(metric["name"]) + " for " + str(aliasName))
                        else:
                            self.logger.debug("Metric not in json " + str(metric["name"]) + " for " + str(aliasName))
                else:
                    self.logger.info("No measurements found")

                if self.capture_partitions is True:

                    # get the partition names.
                    partitionList = self.query_partition_info(node_name)

                    # get disk usage. The data seems to always return none but left the code in if the metric starts producing data...
                    for partition in partitionList:
                        partitionData = self.read("/groups/" + str(self.group_id) + "/processes/" + node_name + "/disks/" + partition + "/measurements?granularity=PT1M&period=PT2M&itemsPerPage=500")

                        if partitionData is not None and "measurements" in partitionData:
                            for metric in partitionData["measurements"]:
                                if str(metric["name"]) in self.args['metrics'].keys():
                                    parsed_metrics = self.parse_metrics(metric["dataPoints"])
                                    if parsed_metrics is not None:
                                        source = self.args['metrics'][str(metric['name'])]
                                                              
                                        #self.logger.info("Writing metric " + str(metric["name"]) + ":" + str(parsed_metrics))
                                        device_temp.absolute(key=str(metric["name"]), value=str(parsed_metrics), dimensions={"disk" : partition})
                                    #else:
                                    #    self.logger.info("MISSING: " + str(metric["name"]))
                                #else:
                                #    self.logger.info("metric not in json " + str(metric["name"]))
                        else:
                            self.logger.info("No measurements found")
                            #self.logger.info(data)

                # check if any events have been raised against this host
                self.reportProcessEvent(eventList["hosts"], hostname, device_temp)


    # convert GB/hour to MB/min
    def convertToMBPerMinute(self, valueToConvert):
        try:
            localVal = float(valueToConvert)
            return localVal/60*1024
        except ValueError:
            return 0


    # format is mongodb://alias1:port,alias2:27017,alias3:27017
    def getNodeNames(self, cluster_info):
        if "mongoURI" in cluster_info:
            return cluster_info["mongoURI"].split("mongodb://")[1].split(",")
        else:
            self.logger.info("Cannot find mongoURI so no node info")
            return []


    # returns a list of partitions for a given host
    def query_partition_info(self, node_name):
        try:
            partitionList = []
            partition_info = self.read("/groups/" + str(self.group_id) + "/processes/" + node_name + "/disks")
            #self.logger.info(partition_info)
            if partition_info is not None:
                for partition in partition_info["results"]:
                    partitionList.append(partition["partitionName"])
            return partitionList
        except Exception as e:
            self.logger.info("Couldn't retrieve partition info:" + str(e))

 


    # generic HTTP call. Returns a json object with the output or None
    def read(self, path):
        #
        # PAGINATION not implemented as the time window is so small but consider as future improvement

        try:
            response = requests.get(self.api_url + path, auth=HTTPDigestAuth(str(self.api_user), str(self.api_token)), proxies=self.proxies, verify=self.verify, timeout = 5)
            if response.status_code == 200:
                self.logger.debug(response.json())
                return response.json()
            else:
                self.logger.info("Read failed with http error " + str(response.status_code) + " for " + path )
                return None

        except Exception as e:
            self.logger.error("Cannot call API: " + str(e)) 
            self.logger.info ('Error on line {}'.format(sys.exc_info()[-1].tb_lineno))
            return None
            

    # one metric uses the source tag because the metric needs to be converted before saving. 
    # it is because the value in returned in GB/Hour which isn't available in DT so converted to MB/Minute
    # the source tag references a function to handle the convertion
    def initialize_metrics(self):
        json_config_metrics = self.json_config['metrics']
        try:
            results = {}
            for metric in json_config_metrics:
                if "source" in metric['timeseries']:
                    source = metric['timeseries']['source']
                else:
                    source = ""
                result = {metric['timeseries']['key']: source}
                results.update(result)
            return results
        except Exception as ex:
            self.logger.error("cannot initialise metrics: " + str(ex))
            return



    # Api returns a series of datapoints broken down by minute. Get the most recent one which is at the end of the array
    # This helps for clusters which return data really slowly. We query the last 2 minutes and take the most recent data point.
    # If the last point is null try the previous one
    # There seems to be a bug where some of the metrics always return null on the first data point.
    def parse_metrics(self, dataPoints):
        numRecords = len(dataPoints)
        if numRecords == 0:
            return None

        index = numRecords -1
        while index >= 0:
            if dataPoints[index]["value"] is None:
                index = index -1 
            else:
                return dataPoints[index]["value"]
        return None



    # scan the events and categorize them: cluster, host or organisation
    def processEvent(self, event_info):
        hostList = defaultdict(set)
        clusterList = defaultdict(set)
        organisationList = []
        if event_info is not None and "results" in event_info:
            for event in event_info["results"]:
                if event["eventTypeName"] in self.GROUP_EVENT_TYPES:
                        # check the scope of the event
                    if "hostname" in event:
                        hostList[event["hostname"]].add(event["eventTypeName"])
                    elif "clusterName" in event:
                        clusterList[event["clusterName"]].add(event["eventTypeName"])
                    else:
                        organisationList.append(event["eventTypeName"])

        return {"hosts" : hostList, "clusters" : clusterList, "organisation" : organisationList}


    # Checks for cluster events and raise them against the custom device group
    # In the very unlikely event of having multiple events, put each one under the properties
    def reportClusterEvents(self, clusterList, clusterName, customDeviceGroup):
        for cluster in clusterList:
            if clusterName == cluster:
                myProps = dict()
                i = 1
                for alert in clusterList[clusterName]:
                    myProps["Event " + str(i)] = alert
                    i += 1
                self.results_builder.report_error_event("Cluster event for " + clusterName, "MongoDB Atlas cluster event", properties=myProps, entity_selector=customDeviceGroup.selector)
                break

                     
    # Checks for host events and raise them against the custom device
    # In the very unlikely event of having multiple events, put each one under the properties    
    def reportProcessEvent(self, hostList, hostname, device_temp):
        for host in hostList:
            if hostname == host:
                myProps = dict()
                i = 1
                for alert in hostList[hostname]:
                    myProps["Event " + str(i)] = alert
                    i += 1
                device_temp.report_error_event("Host event for " + hostname, "MongoDB Atlas host event", properties=myProps)
                break